# Written by Bram Cohen
# Modified by Cameron Dale
# see LICENSE.txt for license information
#
# $Id: download_bt1.py 65 2007-05-27 00:41:11Z camrdale-guest $

"""Manage a single download.

@type DEBUG: C{boolean}
@var DEBUG: whether to print debugging information
@type defaults: C{list} of C{tuple}
@var defaults: the default configuration variables, including descriptions
@type argslistheader: C{string}
@var argslistheader: the header to print before the default config

"""

from zurllib import urlopen
from urlparse import urlparse
from BT1.btformats import check_message
from BT1.Choker import Choker
from BT1.Storage import Storage
from BT1.StorageWrapper import StorageWrapper
from BT1.FileSelector import FileSelector
from BT1.Uploader import Upload
from BT1.Downloader import Downloader
from BT1.HTTPDownloader import HTTPDownloader
from BT1.Connecter import Connecter
from RateLimiter import RateLimiter
from BT1.Encrypter import Encoder
from RawServer import RawServer, autodetect_ipv6, autodetect_socket_style
from BT1.Rerequester import Rerequester
from BT1.DownloaderFeedback import DownloaderFeedback
from RateMeasure import RateMeasure
from CurrentRateMeasure import Measure
from BT1.PiecePicker import PiecePicker
from BT1.Statistics import Statistics
from ConfigDir import ConfigDir
from bencode import bencode, bdecode
from natpunch import UPnP_test
from sha import sha
from os import path, makedirs, listdir, system
from parseargs import parseargs, formatDefinitions, defaultargs
from socket import error as socketerror
from random import seed
from threading import Thread, Event
from clock import clock
from BTcrypto import CRYPTO_OK
from __init__ import createPeerID
from BT1.makemetafile import uniconvertl, uniconvert
from os.path import split
from gzip import GzipFile
from StringIO import StringIO
import binascii
try:
    from sys import getfilesystemencoding
    ENCODING = getfilesystemencoding()
except:
    from sys import getdefaultencoding
    ENCODING = getdefaultencoding()

try:
    True
except:
    True = 1
    False = 0

DEBUG = True

defaults = [
    ('max_uploads', 7,
        "the maximum number of uploads to allow at once."),
    ('keepalive_interval', 120.0,
        'number of seconds to pause between sending keepalives'),
    ('download_slice_size', 2 ** 14,
        "How many bytes to query for per request."),
    ('upload_unit_size', 1460,
        "when limiting upload rate, how many bytes to send at a time"),
    ('request_backlog', 10,
        "maximum number of requests to keep in a single pipe at once."),
    ('max_message_length', 2 ** 23,
        "maximum length prefix encoding you'll accept over the wire - larger values get the connection dropped."),
    ('ip', '',
        "ip to report you have to the tracker."),
    ('minport', 10000, 'minimum port to listen on, counts up if unavailable'),
    ('maxport', 60000, 'maximum port to listen on'),
    ('random_port', 1, 'whether to choose randomly inside the port range ' +
        'instead of counting up linearly'),
    ('responsefile', '',
        'file the server response was stored in, alternative to url'),
    ('url', '',
        'url to get file from, alternative to responsefile'),
    ('crypto_allowed', int(CRYPTO_OK),
        'whether to allow the client to accept encrypted connections'),
    ('crypto_only', 0,
        'whether to only create or allow encrypted connections'),
    ('crypto_stealth', 0,
        'whether to prevent all non-encrypted connection attempts; ' +
        'will result in an effectively firewalled state on older trackers'),
    ('selector_enabled', 1,
        'whether to enable the file selector and fast resume function'),
    ('expire_cache_data', 10,
        'the number of days after which you wish to expire old cache data ' +
        '(0 = disabled)'),
    ('priority', '',
        'a list of file priorities separated by commas, must be one per file, ' +
        '0 = highest, 1 = normal, 2 = lowest, -1 = download disabled'),
    ('saveas', '',
        'local file name to save the file as, null indicates query user'),
    ('timeout', 300.0,
        'time to wait between closing sockets which nothing has been received on'),
    ('timeout_check_interval', 60.0,
        'time to wait between checking if any connections have timed out'),
    ('max_slice_length', 2 ** 17,
        "maximum length slice to send to peers, larger requests are ignored"),
    ('max_rate_period', 20.0,
        "maximum amount of time to guess the current rate estimate represents"),
    ('bind', '', 
        'comma-separated list of ips/hostnames to bind to locally'),
#    ('ipv6_enabled', autodetect_ipv6(),
    ('ipv6_enabled', 0,
         'allow the client to connect to peers via IPv6'),
    ('ipv6_binds_v4', autodetect_socket_style(),
        "set if an IPv6 server socket won't also field IPv4 connections"),
    ('upnp_nat_access', 1,
        'attempt to autoconfigure a UPnP router to forward a server port ' +
        '(0 = disabled, 1 = mode 1 [fast], 2 = mode 2 [slow])'),
    ('upload_rate_fudge', 5.0, 
        'time equivalent of writing to kernel-level TCP buffer, for rate adjustment'),
    ('tcp_ack_fudge', 0.03,
        'how much TCP ACK download overhead to add to upload rate calculations ' +
        '(0 = disabled)'),
    ('display_interval', .5,
        'time between updates of displayed information'),
    ('rerequest_interval', 5 * 60,
        'time to wait between requesting more peers'),
    ('min_peers', 20, 
        'minimum number of peers to not do rerequesting'),
    ('http_timeout', 60, 
        'number of seconds to wait before assuming that an http connection has timed out'),
    ('max_initiate', 40,
        'number of peers at which to stop initiating new connections'),
    ('check_hashes', 1,
        'whether to check hashes on disk'),
    ('max_upload_rate', 0,
        'maximum kB/s to upload at (0 = no limit, -1 = automatic)'),
    ('max_download_rate', 0,
        'maximum kB/s to download at (0 = no limit)'),
# Remove allocation controls for now (must pre-allocate)
#    ('alloc_type', 'pre-allocate',
#        'allocation type (may be normal, background, pre-allocate or sparse)'),
#    ('alloc_rate', 2.0,
#        'rate (in MiB/s) to allocate space at using background allocation'),
    ('buffer_reads', 1,
        'whether to buffer disk reads'),
    ('write_buffer_size', 4,
        'the maximum amount of space to use for buffering disk writes ' +
        '(in megabytes, 0 = disabled)'),
    ('breakup_seed_bitfield', 1,
        'sends an incomplete bitfield and then fills with have messages, '
        'in order to get around stupid ISP manipulation'),
    ('snub_time', 30.0,
        "seconds to wait for data to come in over a connection before assuming it's semi-permanently choked"),
    ('spew', 0,
        "whether to display diagnostic info to stdout"),
    ('rarest_first_cutoff', 2,
        "number of downloads at which to switch from random to rarest first"),
    ('rarest_first_priority_cutoff', 5,
        'the number of peers which need to have a piece before other partials take priority over rarest first'),
    ('min_uploads', 4,
        "the number of uploads to fill out to with extra optimistic unchokes"),
    ('max_files_open', 50,
        'the maximum number of files to keep open at a time, 0 means no limit'),
    ('round_robin_period', 30,
        "the number of seconds between the client's switching upload targets"),
    ('super_seeder', 0,
        "whether to use special upload-efficiency-maximizing routines (only for dedicated seeds)"),
    ('security', 1,
        "whether to enable extra security features intended to prevent abuse"),
    ('max_connections', 0,
        "the absolute maximum number of peers to connect with (0 = no limit)"),
    ('auto_kick', 1,
        "whether to allow the client to automatically kick/ban peers that send bad data"),
    ('double_check', 1,
        "whether to double-check data being written to the disk for errors (may increase CPU load)"),
    ('triple_check', 0,
        "whether to thoroughly check data being written to the disk (may slow disk access)"),
    ('lock_files', 1,
        "whether to lock files the client is working with"),
    ('lock_while_reading', 0,
        "whether to lock access to files being read"),
    ('auto_flush', 0,
        "minutes between automatic flushes to disk (0 = disabled)"),
    ('dedicated_seed_id', '',
        "code to send to tracker identifying as a dedicated seed"),
    ('status_to_download', 0,
        'determines which packages to download based on /var/lib/dpkg/status ' +
        '(0 = disabled [download all or use --priority], 1 = download updated versions of installed packages, ' +
         '2 = download all installed packages)'),
    ]

argslistheader = 'Arguments are:\n\n'


def _failfunc(x):
    """Default function to use for printing error messages.
    
    @type x: C{string}
    @param x: the error message to print
    
    """
    
    print x

def download(params, filefunc, statusfunc, finfunc, errorfunc, doneflag, cols,
             pathFunc = None, presets = {}, exchandler = None,
             failed = _failfunc, paramfunc = None):
    """The old-style downloader (no longer used).
    
    @type params: unknown
    @param params: unknown
    @type filefunc: unknown
    @param filefunc: unknown
    @type statusfunc: unknown
    @param statusfunc: unknown
    @type finfunc: unknown
    @param finfunc: unknown
    @type errorfunc: unknown
    @param errorfunc: unknown
    @type doneflag: unknown
    @param doneflag: unknown
    @type cols: unknown
    @param cols: unknown
    @type pathFunc: unknown
    @param pathFunc: unknown
    @type presets: unknown
    @param presets: unknown
    @type exchandler: unknown
    @param exchandler: unknown
    @type failed: unknown
    @param failed: unknown
    @type paramfunc: unknown
    @param paramfunc: unknown
    
    """

    try:
        config = parse_params(params, presets)
    except ValueError, e:
        failed('error: ' + str(e) + '\nrun with no args for parameter explanations')
        return
    if not config:
        errorfunc(get_usage())
        return
    
    myid = createPeerID()
    seed(myid)

    rawserver = RawServer(doneflag, config['timeout_check_interval'],
                          config['timeout'], ipv6_enable = config['ipv6_enabled'],
                          failfunc = failed, errorfunc = exchandler)

    upnp_type = UPnP_test(config['upnp_nat_access'])
    try:
        listen_port = rawserver.find_and_bind(config['minport'], config['maxport'],
                        config['bind'], ipv6_socket_style = config['ipv6_binds_v4'],
                        upnp = upnp_type, randomizer = config['random_port'])
    except socketerror, e:
        failed("Couldn't listen - " + str(e))
        return

    (response, status_priority) = get_response(config['responsefile'], config['url'], failed)
    if not response:
        return

    infohash = sha(bencode(response['info'])).digest()

    d = BT1Download(statusfunc, finfunc, errorfunc, exchandler, doneflag,
                    config, response, infohash, myid, rawserver, listen_port, status_priority)

    if not d.saveAs(filefunc):
        return

    if pathFunc:
        pathFunc(d.getFilename())

    hashcheck = d.initFiles(old_style = True)
    if not hashcheck:
        return
    if not hashcheck():
        return
    if not d.startEngine():
        return
    d.startRerequester()
    d.autoStats()

    statusfunc(activity = 'connecting to peers')

    if paramfunc:
        paramfunc({ 'max_upload_rate' : d.setUploadRate,  # change_max_upload_rate(<int KiB/sec>)
                    'max_uploads': d.setConns, # change_max_uploads(<int max uploads>)
                    'listen_port' : listen_port, # int
                    'peer_id' : myid, # string
                    'info_hash' : infohash, # string
                    'start_connection' : d._startConnection, # start_connection((<string ip>, <int port>), <peer id>)
                    })
        
    rawserver.listen_forever(d.getPortHandler())
    
    d.shutdown()


def parse_params(params, presets = {}):
    """Parse the command-line parameters.
    
    @type params: C{list} of C{string}
    @param params: the command-line parameters
    @type presets: C{dictionary}
    @param presets: the preset values to use (optional)
    @rtype: C{dictionary}
    @return: the configuration variables
    @raise ValueError: if the parameters are not properly specified
    
    """
    
    if len(params) == 0:
        return None
    config, args = parseargs(params, defaults, 0, 1, presets = presets)
    if args:
        if config['responsefile'] or config['url']:
            raise ValueError,'must have responsefile or url as arg or parameter, not both'
        if path.isfile(args[0]):
            config['responsefile'] = args[0]
        else:
            try:
                urlparse(args[0])
            except:
                raise ValueError, 'bad filename or url'
            config['url'] = args[0]
    elif (config['responsefile'] == '') == (config['url'] == ''):
        raise ValueError, 'need responsefile or url, must have one, cannot have both'
    return config


def get_usage(defaults = defaults, cols = 100, presets = {}):
    """Print the usage information for the program.
    
    @type defaults: C{list} of C{tuple}
    @param defaults: the default configuration variables
        (optional, default is to use L{defaults})
    @type cols: C{int}
    @param cols: the width of the print out (optional, default is 100)
    @type presets: C{dictionary}
    @param presets: the preset values to use (optional)
    
    """
    
    return (argslistheader + formatDefinitions(defaults, cols, presets))


def get_response(file, url, status_to_download, errorfunc):
    """Get the response data from a metainfo or Packages file.
    
    First checks to see if the data is in the Packages file format, and
    returns the extracted response data if it is. Otherwise, assumes it is
    a metainfo file and tries to get the response data from it.
    
    @type file: C{string}
    @param file: the file name to use, or None to indicate that the url is
        to be used
    @type url: C{string}
    @param url: the URL to download the metainfo file from
    @type status_to_download: C{int}
    @param status_to_download: determines which packages to download based on 
        /var/lib/dpkg/status (0 = disabled [download all or use --priority], 
        1 = download updated versions of installed packages,
        2 = download all installed packages)
    @type errorfunc: C{function}
    @param errorfunc: the function to use to print any error messages
    @rtype: C{dictionary}
    @return: the metainfo data
    
    """
    
    (response, priority) = get_packages(file, url, status_to_download)
    if response:
        try:
            check_message(response)
        except ValueError, e:
            errorfunc("got bad file info - " + str(e))
            return (None, None)
        return (response, priority)
    try:
        if file:
            h = open(file, 'rb')
            try:
                line = h.read(10)   # quick test to see if responsefile contains a dict
                front,garbage = line.split(':',1)
                assert front[0] == 'd'
                int(front[1:])
            except:
                errorfunc(file+' is not a valid responsefile')
                return (None, None)
            try:
                h.seek(0)
            except:
                try:
                    h.close()
                except:
                    pass
                h = open(file, 'rb')
        else:
            try:
                h = urlopen(url)
            except:
                errorfunc(url+' bad url')
                return (None, None)
        response = h.read()
    
    except IOError, e:
        errorfunc('problem getting response info - ' + str(e))
        return (None, None)
    try:    
        h.close()
    except:
        pass
    try:
        try:
            response = bdecode(response)
        except:
            errorfunc("warning: bad data in responsefile")
            response = bdecode(response, sloppy=1)
        check_message(response)
    except ValueError, e:
        errorfunc("got bad file info - " + str(e))
        return (None, None)

    return (response, None)

def get_packages(file, url, status_to_download):
    """Extract the response data from a Packages file.
    
    @type file: C{string}
    @param file: the file name to use, or None to indicate that the url is
        to be used
    @type url: C{string}
    @param url: the URL to download the metainfo file from
    @type status_to_download: C{int}
    @param status_to_download: determines which packages to download based on 
        /var/lib/dpkg/status (0 = disabled [download all or use --priority], 
        1 = download updated versions of installed packages,
        2 = download all installed packages)
    @rtype: C{dictionary}
    @return: the metainfo data
    
    """

    encoding = None
#    if params.has_key('filesystem_encoding'):
#        encoding = params['filesystem_encoding']
    if not encoding:
        encoding = ENCODING
    if not encoding:
        encoding = 'ascii'
        
    get_priority = False
    installed_versions = {}
    priority = None
    if status_to_download > 0:
        try:
            status = open("/var/lib/dpkg/status")
    
            p = [None, None, None]
            for line in status:
                line = line.rstrip()
                if line == "":
                    if (p[0] and p[1] and p[2]):
                        if (p[2] == "install ok installed"):
                            installed_versions[p[0]] = p[1]
                        p = [None, None, None]
                if line[:8] == "Package:":
                    p[0] = line[9:]
                if line[:8] == "Version:":
                    p[1] = line[9:]
                if line[:7] == "Status:":
                    p[2] = line[8:]
        except:
            installed_versions = {}

        try:
            status.close()
        except:
            pass
        
        if installed_versions:
            get_priority = True
            priority = {}

    try:
        if file:
            name = "dt_" + split(file)[1]
            h = open(file, 'r')
            try:
                for line in h:   # quick test to see if packages file is correct
                    if len(line.strip()) > 0:
                        break
                assert line[:8] == "Package:"
            except:
#                errorfunc(file+' is not a valid Packages file')
                return (None, None)
            try:
                h.seek(0)
            except:
                try:
                    h.close()
                except:
                    pass
                h = open(file, 'r')
        else:
            urlp = urlparse(url)
            name = "dt_" + urlp[1] + urlp[2].replace('/','_')
            try:
                h = urlopen(url)
                file = url
                data = h.read()
                assert len(data) > 8
                if url[-3:] == ".gz":
                    name = "dt_" + urlp[1] + urlp[2][:-3].replace('/','_')
                    if data[:8] != "Package:":
                        compressed = StringIO(data)
                        f = GzipFile(fileobj = compressed)
                        data = f.read()
                assert data[:8] == "Package:"
                h = data.split('\n')
            except:
#                errorfunc(url+' bad url')
                return (None, None)

        pieces = []
        lengths = []
        fs = []
        
        p = [None, None, None, None, None]
        for line in h:
            line = line.rstrip()
            if line == "":
                if (p[0] and p[1] and p[2]):
                    file_path = []
                    temp = p[1]
                    while p[1]:
                        p[1],d = split(p[1])
                        file_path.insert(0,d)
                    fs.append({'length': p[0], 'path': uniconvertl(file_path, encoding)})
                    lengths.append(p[0])
                    pieces.append(p[2])
                    if (get_priority and p[3] and installed_versions.has_key(p[3])):
                        if status_to_download == 2:
                            priority[temp] = 1
                        elif (status_to_download == 1 and p[4]):
                            ret = system('dpkg --compare-versions "' + installed_versions[p[3]] + '" lt "' + p[4] + '"')
                            if ret == 0:
                                priority[temp] = 1
                p = [None, None, None, None, None]
            if line[:9] == "Filename:":
                p[1] = line[10:]
            if line[:5] == "Size:":
                p[0] = long(line[6:])
            if line[:5] == "SHA1:":
                p[2] = binascii.a2b_hex(line[6:])
            if line[:8] == "Package:":
                p[3] = line[9:]
            if line[:8] == "Version:":
                p[4] = line[9:]
        
        response = {'info': {'pieces': ''.join(pieces),
            'piece lengths': lengths, 'files': fs },
            'announce': 'http://dttracker.debian.net:6969/announce', 
            'name': uniconvert(name, encoding) }
    
    except IOError, e:
#        errorfunc('problem getting Packages info - ' + str(e))
        return (None, None)
    try:    
        h.close()
    except:
        pass

    return (response, priority)


class BT1Download:    
    """Manage a single download.
    
    @type statusfunc: C{method}
    @ivar statusfunc: the method to call to print status updates
    @type finfunc: C{method}
    @ivar finfunc: the method to call when the download is completed
    @type errorfunc: C{method}
    @ivar errorfunc: the method to call when an error occurs
    @type excfunc: C{method}
    @ivar excfunc: the method to call when an exception occurs
    @type doneflag: C{threading.Event}
    @ivar doneflag: the flag that indicates when the program is to be shutdown
    @type config: C{dictionary}
    @ivar config: the configuration variables
    @type response: C{dictionary}
    @ivar response: the response data from the metainfo file
    @type infohash: C{string}
    @ivar infohash: the hash of the info from the response data
    @type myid: C{string}
    @ivar myid: the peer ID to use
    @type rawserver: L{Rawserver.Rawserver}
    @ivar rawserver: the server controlling the program
    @type port: C{int}
    @ivar port: the port being listened to
    @type info: C{dictionary}
    @ivar info: the info data from the response data
    @type pieces: C{list} of C{string}
    @ivar pieces: the hashes of the pieces
    @type piece_lengths: C{list} of C{int}
    @ivar piece_lengths: the lengths of the pieces
    @type len_pieces: C{int}
    @ivar len_pieces: the number of pieces
    @type status_priority: C{dictionary}
    @ivar status_priority: the file priorities from the dpkg/status file
    @type argslistheader: C{string}
    @ivar argslistheader: the header to print before the default config
    @type unpauseflag: C{threading.Event}
    @ivar unpauseflag: the flag to unset to pause the download
    @type downloader: L{BT1.Downloader.Downloader}
    @ivar downloader: the Downloader instance
    @type storagewrapper: L{BT1.StorageWrapper.StorageWrapper}
    @ivar storagewrapper: the StorageWrapper instance
    @type fileselector: L{BT1.FileSelector.FileSelector}
    @ivar fileselector: the FileSelector instance
    @type super_seeding_active: C{boolean}
    @ivar super_seeding_active: whether the download is in super-seed mode
    @type filedatflag: C{threading.Event}
    @ivar filedatflag: unknown
    @type spewflag: C{threading.Event}
    @ivar spewflag: unknown
    @type superseedflag: C{threading.Event}
    @ivar superseedflag: unknown
    @type whenpaused: unknown
    @ivar whenpaused: unknown
    @type finflag: C{threading.Event}
    @ivar finflag: unknown
    @type rerequest: unknown
    @ivar rerequest: unknown
    @type tcp_ack_fudge: C{float}
    @ivar tcp_ack_fudge: the fraction of TCP ACK download overhead to add to 
        upload rate calculations
    @type selector_enabled: C{boolean}
    @ivar selector_enabled: whether to enable the file selector and fast resume function
    @type appdataobj: L{ConfigDir.ConfigDir}
    @ivar appdataobj: the configuration and cache directory manager
    @type excflag: C{threading.Event}
    @ivar excflag: unknown
    @type failed: unknown
    @ivar failed: unknown
    @type checking: unknown
    @ivar checking: unknown
    @type started: unknown
    @ivar started: unknown
    @type picker: L{BT1.PiecePicker.PiecePicker}
    @ivar picker: the PiecePicker instance
    @type choker: L{BT1.Choker.Choker}
    @ivar choker: the Choker instance
    @type filename: C{string}
    @ivar filename: the save location
    @type files: C{list} of (C{string}, C{long})
    @ivar files: the full file names and lengths of all the files in the download
    @type datalength: C{long}
    @ivar datalength: the total length of the download
    @type priority: unknown
    @ivar priority: unknown
    @type storage: unknown
    @ivar storage: unknown
    @type upmeasure: unknown
    @ivar upmeasure: unknown
    @type downmeasure: unknown
    @ivar downmeasure: unknown
    @type ratelimiter: unknown
    @ivar ratelimiter: unknown
    @type ratemeasure: unknown
    @ivar ratemeasure: unknown
    @type ratemeasure_datarejected: unknown
    @ivar ratemeasure_datarejected: unknown
    @type connecter: unknown
    @ivar connecter: unknown
    @type encoder: unknown
    @ivar encoder: unknown
    @type encoder_ban: unknown
    @ivar encoder_ban: unknown
    @type httpdownloader: unknown
    @ivar httpdownloader: unknown
    @type statistics: unknown
    @ivar statistics: unknown
    
    """
    
    def __init__(self, statusfunc, finfunc, errorfunc, excfunc, doneflag,
                 config, response, infohash, id, rawserver, port,
                 status_priority = None, appdataobj = None):
        """Initialize the instance.
        
        @type statusfunc: C{method}
        @param statusfunc: the method to call to print status updates
        @type finfunc: C{method}
        @param finfunc: the method to call when the download is completed
        @type errorfunc: C{method}
        @param errorfunc: the method to call when an error occurs
        @type excfunc: C{method}
        @param excfunc: the method to call when an exception occurs
        @type doneflag: C{threading.Event}
        @param doneflag: the flag that indicates when the program is to be shutdown
        @type config: C{dictionary}
        @param config: the configuration variables
        @type response: C{dictionary}
        @param response: the response data from the metainfo file
        @type infohash: C{string}
        @param infohash: the hash of the info from the response data
        @type id: C{string}
        @param id: the peer ID to use
        @type rawserver: L{Rawserver.Rawserver}
        @param rawserver: the server controlling the program
        @type port: C{int}
        @param port: the port being listened to
        @type status_priority: C{dictionary}
        @param status_priority: the file priorities, keys are file names, 
            values are the priority to use (optional, defaults to download all)
        @type appdataobj: L{ConfigDir.ConfigDir}
        @param appdataobj: the configuration and cache directory manager
        
        """
        
        self.statusfunc = statusfunc
        self.finfunc = finfunc
        self.errorfunc = errorfunc
        self.excfunc = excfunc
        self.doneflag = doneflag
        self.config = config
        self.response = response
        self.infohash = infohash
        self.myid = id
        self.rawserver = rawserver
        self.port = port
        
        self.info = self.response['info']
        self.pieces = [self.info['pieces'][x:x+20]
                       for x in xrange(0, len(self.info['pieces']), 20)]
        self.piece_lengths = self.info['piece lengths']
        self.len_pieces = len(self.pieces)
        self.status_priority = status_priority
        self.argslistheader = argslistheader
        self.unpauseflag = Event()
        self.unpauseflag.set()
        self.downloader = None
        self.storagewrapper = None
        self.fileselector = None
        self.super_seeding_active = False
        self.filedatflag = Event()
        self.spewflag = Event()
        self.superseedflag = Event()
        self.whenpaused = None
        self.finflag = Event()
        self.rerequest = None
        self.tcp_ack_fudge = config['tcp_ack_fudge']

        self.selector_enabled = config['selector_enabled']
        if appdataobj:
            self.appdataobj = appdataobj
        elif self.selector_enabled:
            self.appdataobj = ConfigDir()
            self.appdataobj.deleteOldCacheData( config['expire_cache_data'],
                                                [self.infohash] )

        self.excflag = self.rawserver.get_exception_flag()
        self.failed = False
        self.checking = False
        self.started = False

        self.picker = PiecePicker(self.len_pieces, config['rarest_first_cutoff'],
                             config['rarest_first_priority_cutoff'])
        self.choker = Choker(config, rawserver.add_task,
                             self.picker, self.finflag.isSet)


    def checkSaveLocation(self, loc):
        """Check whether the download location exists.
        
        For multiple files, returns true if a single one exists.
        
        @type loc: C{string}
        @param loc: the save location to check
        @rtype: C{boolean}
        @return: whether the location exists.
        
        """
        
        if self.info.has_key('length'):
            return path.exists(loc)
        for x in self.info['files']:
            if path.exists(path.join(loc, x['path'][0])):
                return True
        return False
                

    def saveAs(self, filefunc, pathfunc = None):
        """Initialize the location to save the download to.
        
        @type filefunc: C{method}
        @param filefunc: the method to call to get the location to save the 
            download
        @type pathfunc: C{method}
        @param pathfunc: the method to call to alert the UI to any possible 
            change in the download location
        @rtype: C{string}
        @return: the location to save to
        
        """
        
        try:
            def make(f, forcedir = False):
                """Create the parent directories of a file.
                
                @type f: C{string}
                @param f: the file name
                @type forcedir: C{boolean}
                @param forcedir: set to True if f is a directory name
                    (optional, defaults to False)
                
                """
                
                if not forcedir:
                    f = path.split(f)[0]
                if f != '' and not path.exists(f):
                    makedirs(f)

            if self.info.has_key('length'):
                file_length = self.info['length']
                file = filefunc(self.response['name'], file_length,
                                self.config['saveas'], False)
                if file is None:
                    return None
                make(file)
                files = [(file, file_length)]
            else:
                file_length = 0L
                for x in self.info['files']:
                    file_length += x['length']
                file = filefunc(self.response['name'], file_length,
                                self.config['saveas'], True)
                if file is None:
                    return None

                # if this path exists, and no files from the info dict exist, we assume it's a new download and 
                # the user wants to create a new directory with the default name
                existing = 0
                if path.exists(file):
                    if not path.isdir(file):
                        self.errorfunc(file + 'is not a dir')
                        return None
                    if len(listdir(file)) > 0:  # if it's not empty
                        for x in self.info['files']:
                            if path.exists(path.join(file, x['path'][0])):
                                existing = 1
                        if not existing:
                            file = path.join(file, self.response['name'])
                            if path.exists(file) and not path.isdir(file):
                                if file[-8:] == '.dtorrent':
                                    file = file[:-8]
                                if path.exists(file) and not path.isdir(file):
                                    self.errorfunc("Can't create dir - " + self.response['name'])
                                    return None
                make(file, True)

                # alert the UI to any possible change in path
                if pathfunc != None:
                    pathfunc(file)

                files = []
                for x in self.info['files']:
                    n = file
                    for i in x['path']:
                        n = path.join(n, i)
                    files.append((n, x['length']))
                    #Move directory creation to Storage
                    #make(n)
        except OSError, e:
            self.errorfunc("Couldn't allocate dir - " + str(e))
            return None

        self.filename = file
        self.files = files
        self.datalength = file_length

        return file
    

    def getFilename(self):
        """Get the download location.
        
        @rtype: C{string}
        @return: the download location
        
        """
        
        return self.filename


    def _finished(self):
        """Finish the download.
        
        Set the files to read-only, update the Choker for seeding, stop the 
        piece requester.
        
        """
        
        self.finflag.set()
        try:
            self.storage.set_readonly()
        except (IOError, OSError), e:
            self.errorfunc('trouble setting readonly at end - ' + str(e))
        if self.superseedflag.isSet():
            self._set_super_seed()
        self.choker.set_round_robin_period(
            max( self.config['round_robin_period'],
                 self.config['round_robin_period'] *
                           (float(self.datalength) / len(self.piece_lengths)) / 200000 ) )
        self.rerequest_complete()
        self.finfunc()

    def _data_flunked(self, amount, index):
        """Process a failed hash check on a piece.
        
        @type amount: C{int}
        @param amount: the amount of failed data
        @type index: C{int}
        @param index: the piece that failed
        
        """
        
        self.ratemeasure_datarejected(amount)
        if not self.doneflag.isSet():
            self.errorfunc('piece %d failed hash check, re-downloading it' % index)

    def _failed(self, reason):
        """Stop the failed download.
        
        @type reason: C{string}
        @param reason: the reason for the failure
        
        """
        
        self.failed = True
        self.doneflag.set()
        if reason is not None:
            self.errorfunc(reason)
        

    def initFiles(self, old_style = False, statusfunc = None):
        """Initialize the files for the download.
        
        Initialize the priorities, then create the Storage, StorageWrapper, 
        and FileSelector. Then initialize the StorageWrapper.
        
        @type old_style: C{boolean}
        @param old_style: whether to use the old-style StorageWrapper 
            initialization (optional, defaults to False)
        @type statusfunc: C{method}
        @param statusfunc: the method to use to diplay status updates
            (optional, defaults to using L{statusfunc}
        @rtype: C{boolean}
        @return: whether the initialization was successful
        
        """
        
        if self.doneflag.isSet():
            return None
        if not statusfunc:
            statusfunc = self.statusfunc

        disabled_files = None
        if self.selector_enabled:
            if self.status_priority is not None:
                if DEBUG:
                    print 'original priority:', str(self.status_priority)
                if len(self.status_priority) == 0:
                    self._failed('Nothing to download')
                    if self.doneflag.isSet():
                        return None
                updated_priority = {}
                for k, v in self.status_priority.items():
                    updated_priority[path.join(self.filename,k)] = v
                if DEBUG:
                    print 'updated priority:', str(updated_priority)
                self.status_priority = []
                for file, length in self.files:
                    self.status_priority.append(updated_priority.get(file, -1))
                if DEBUG:
                    print 'final priority:', str(self.status_priority)
            self.priority = self.config['priority']
            if self.priority:
                try:
                    self.priority = self.priority.split(',')
                    assert len(self.priority) == len(self.files)
                    self.priority = [int(p) for p in self.priority]
                    for p in self.priority:
                        assert p >= -1
                        assert p <= 2
                except:
                    self.errorfunc('bad priority list given, ignored')
                    self.priority = self.status_priority
            else:
                self.priority = self.status_priority

            data = self.appdataobj.getTorrentData(self.infohash)
            try:
                d = data['resume data']['priority']
                assert len(d) == len(self.files)
                disabled_files = [x == -1 for x in d]
            except:
                try:
                    disabled_files = [x == -1 for x in self.priority]
                except:
                    pass

        try:
            try:
                self.storage = Storage(self.files, self.piece_lengths,
                                       self.doneflag, self.config, disabled_files)
            except IOError, e:
                self.errorfunc('trouble accessing files - ' + str(e))
                return None
            if self.doneflag.isSet():
                return None

            self.storagewrapper = StorageWrapper(self.storage, self.config['download_slice_size'],
                self.pieces, self.piece_lengths, self.datalength, self._finished, self._failed,
                statusfunc, self.doneflag, self.config['check_hashes'],
                self._data_flunked, self.rawserver.add_task,
                self.config, self.unpauseflag)
            
        except ValueError, e:
            self._failed('bad data - ' + str(e))
        except IOError, e:
            self._failed('IOError - ' + str(e))
        if self.doneflag.isSet():
            return None

        if self.selector_enabled:
            self.fileselector = FileSelector(self.files, self.piece_lengths,
                                             self.appdataobj.getPieceDir(self.infohash),
                                             self.storage, self.storagewrapper,
                                             self.rawserver.add_task,
                                             self._failed)
            if data:
                data = data.get('resume data')
                if data:
                    self.fileselector.unpickle(data)

            if self.priority:
                self.fileselector.initialize_priorities_now(self.priority)
                
        self.checking = True
        if old_style:
            return self.storagewrapper.old_style_init()
        return self.storagewrapper.initialize


    def getCachedTorrentData(self):
        """Get the cached torrent data from the cache directory.
        
        @rtype: C{dictionary}
        @return: the bdecoded cached data
        
        """
        
        return self.appdataobj.getTorrentData(self.infohash)


    def _make_upload(self, connection, ratelimiter, totalup):
        """Create a new Upload instance
        
        @type connection: unknown
        @param connection: the connection to upload on
        @type ratelimiter: L{BT1.RateLimiter.RateLimiter}
        @param ratelimiter: the RateLimiter instance to use
        @type totalup: L{CurrentRateMeasure.Measure}
        @param totalup: the Measure instance to use to calculate the total 
            upload rate
        @rtype: L{BT1.Uploader.Upload}
        @return: the new Upload instance
        
        """
        
        return Upload(connection, ratelimiter, totalup,
                      self.choker, self.storagewrapper, self.picker,
                      self.config)

    def _kick_peer(self, connection):
        """Disconnect a peer.
        
        @type connection: unknown
        @param connection: the connection of the peer to disconnect
        
        """
        
        def k(connection = connection):
            """Close a connection.
            
            @type connection: unknown
            @param connection: the connection of the peer to disconnect
                (optional, defaults to the _kick_peer connection)
            
            """
        
            connection.close()
        self.rawserver.add_task(k,0)

    def _ban_peer(self, ip):
        """Ban a peer from the download.
        
        @type ip: C{string}
        @param ip: the IP address of the peer to ban
        
        """
        
        self.encoder_ban(ip)

    def _received_raw_data(self, x):
        if self.tcp_ack_fudge:
            x = int(x*self.tcp_ack_fudge)
            self.ratelimiter.adjust_sent(x)

    def _received_data(self, x):
        self.downmeasure.update_rate(x)
        self.ratemeasure.data_came_in(x)

    def _received_http_data(self, x):
        self.downmeasure.update_rate(x)
        self.ratemeasure.data_came_in(x)
        self.downloader.external_data_received(x)

    def _cancelfunc(self, pieces):
        self.downloader.cancel_piece_download(pieces)
        self.httpdownloader.cancel_piece_download(pieces)
    def _reqmorefunc(self, pieces):
        self.downloader.requeue_piece_download(pieces)

    def startEngine(self, ratelimiter = None, statusfunc = None):
        if self.doneflag.isSet():
            return False
        if not statusfunc:
            statusfunc = self.statusfunc

        self.checking = False

        if not CRYPTO_OK:
            if self.config['crypto_allowed']:
                self.errorfunc('warning - crypto library not installed')
            self.config['crypto_allowed'] = 0
            self.config['crypto_only'] = 0
            self.config['crypto_stealth'] = 0

        for i in xrange(self.len_pieces):
            if self.storagewrapper.do_I_have(i):
                self.picker.complete(i)
        self.upmeasure = Measure(self.config['max_rate_period'],
                            self.config['upload_rate_fudge'])
        self.downmeasure = Measure(self.config['max_rate_period'])

        if ratelimiter:
            self.ratelimiter = ratelimiter
        else:
            self.ratelimiter = RateLimiter(self.rawserver.add_task,
                                           self.config['upload_unit_size'],
                                           self.setConns)
            self.ratelimiter.set_upload_rate(self.config['max_upload_rate'])
        
        self.ratemeasure = RateMeasure()
        self.ratemeasure_datarejected = self.ratemeasure.data_rejected

        self.downloader = Downloader(self.storagewrapper, self.picker,
            self.config['request_backlog'], self.config['max_rate_period'],
            self.len_pieces, self.config['download_slice_size'],
            self._received_data, self.config['snub_time'], self.config['auto_kick'],
            self._kick_peer, self._ban_peer)
        self.downloader.set_download_rate(self.config['max_download_rate'])
        self.connecter = Connecter(self._make_upload, self.downloader, self.choker,
                            self.len_pieces, self.upmeasure, self.config,
                            self.ratelimiter, self.rawserver.add_task)
        self.encoder = Encoder(self.connecter, self.rawserver,
            self.myid, self.config['max_message_length'], self.rawserver.add_task,
            self.config['keepalive_interval'], self.infohash,
            self._received_raw_data, self.config)
        self.encoder_ban = self.encoder.ban

        self.httpdownloader = HTTPDownloader(self.storagewrapper, self.picker,
            self.rawserver, self.finflag, self.errorfunc, self.downloader,
            self.config['max_rate_period'], self.infohash, self._received_http_data,
            self.connecter.got_piece)
        if self.response.has_key('httpseeds') and not self.finflag.isSet():
            for u in self.response['httpseeds']:
                self.httpdownloader.make_download(u)

        if self.selector_enabled:
            self.fileselector.tie_in(self.picker, self._cancelfunc,
                    self._reqmorefunc, self.rerequest_ondownloadmore)
            if self.priority:
                self.fileselector.set_priorities_now(self.priority)
            self.appdataobj.deleteTorrentData(self.infohash)
                                # erase old data once you've started modifying it

        if self.config['super_seeder']:
            self.set_super_seed()

        self.started = True
        return True


    def rerequest_complete(self):
        if self.rerequest:
            self.rerequest.announce(1)

    def rerequest_stopped(self):
        if self.rerequest:
            self.rerequest.announce(2)

    def rerequest_lastfailed(self):
        if self.rerequest:
            return self.rerequest.last_failed
        return False

    def rerequest_ondownloadmore(self):
        if self.rerequest:
            self.rerequest.hit()

    def startRerequester(self, seededfunc = None, force_rapid_update = False):
        if self.response.has_key('announce-list'):
            trackerlist = self.response['announce-list']
        else:
            trackerlist = [[self.response['announce']]]

        self.rerequest = Rerequester(self.port, self.myid, self.infohash, 
            trackerlist, self.config, 
            self.rawserver.add_task, self.rawserver.add_task,
            self.errorfunc, self.excfunc,
            self.encoder.start_connections,
            self.connecter.how_many_connections, 
            self.storagewrapper.get_amount_left, 
            self.upmeasure.get_total, self.downmeasure.get_total,
            self.upmeasure.get_rate, self.downmeasure.get_rate,
            self.doneflag, self.unpauseflag, seededfunc, force_rapid_update )

        self.rerequest.start()


    def _init_stats(self):
        self.statistics = Statistics(self.upmeasure, self.downmeasure,
                    self.connecter, self.httpdownloader, self.ratelimiter,
                    self.rerequest_lastfailed, self.filedatflag)
        if self.info.has_key('files'):
            self.statistics.set_dirstats(self.files, self.piece_lengths)
        if self.config['spew']:
            self.spewflag.set()

    def autoStats(self, displayfunc = None):
        if not displayfunc:
            displayfunc = self.statusfunc

        self._init_stats()
        DownloaderFeedback(self.choker, self.httpdownloader, self.rawserver.add_task,
            self.upmeasure.get_rate, self.downmeasure.get_rate,
            self.ratemeasure, self.storagewrapper.get_stats,
            self.datalength, self.finflag, self.spewflag, self.statistics,
            displayfunc, self.config['display_interval'])

    def startStats(self):
        self._init_stats()
        d = DownloaderFeedback(self.choker, self.httpdownloader, self.rawserver.add_task,
            self.upmeasure.get_rate, self.downmeasure.get_rate,
            self.ratemeasure, self.storagewrapper.get_stats,
            self.datalength, self.finflag, self.spewflag, self.statistics)
        return d.gather


    def getPortHandler(self):
        return self.encoder


    def shutdown(self, torrentdata = {}):
        if self.checking or self.started:
            self.storagewrapper.sync()
            self.storage.close()
            self.rerequest_stopped()
        if self.fileselector and self.started:
            if not self.failed:
                self.fileselector.finish()
                torrentdata['resume data'] = self.fileselector.pickle()
            try:
                self.appdataobj.writeTorrentData(self.infohash,torrentdata)
            except:
                self.appdataobj.deleteTorrentData(self.infohash) # clear it
        return not self.failed and not self.excflag.isSet()
        # if returns false, you may wish to auto-restart the torrent


    def setUploadRate(self, rate):
        try:
            def s(self = self, rate = rate):
                self.config['max_upload_rate'] = rate
                self.ratelimiter.set_upload_rate(rate)
            self.rawserver.add_task(s)
        except AttributeError:
            pass

    def setConns(self, conns, conns2 = None):
        if not conns2:
            conns2 = conns
        try:
            def s(self = self, conns = conns, conns2 = conns2):
                self.config['min_uploads'] = conns
                self.config['max_uploads'] = conns2
                if (conns > 30):
                    self.config['max_initiate'] = conns + 10
            self.rawserver.add_task(s)
        except AttributeError:
            pass
        
    def setDownloadRate(self, rate):
        try:
            def s(self = self, rate = rate):
                self.config['max_download_rate'] = rate
                self.downloader.set_download_rate(rate)
            self.rawserver.add_task(s)
        except AttributeError:
            pass

    def startConnection(self, ip, port, id):
        self.encoder._start_connection((ip, port), id)
      
    def _startConnection(self, ipandport, id):
        self.encoder._start_connection(ipandport, id)
        
    def setInitiate(self, initiate):
        try:
            def s(self = self, initiate = initiate):
                self.config['max_initiate'] = initiate
            self.rawserver.add_task(s)
        except AttributeError:
            pass

    def getConfig(self):
        return self.config

    def getDefaults(self):
        return defaultargs(defaults)

    def getUsageText(self):
        return self.argslistheader

    def reannounce(self, special = None):
        try:
            def r(self = self, special = special):
                if special is None:
                    self.rerequest.announce()
                else:
                    self.rerequest.announce(specialurl = special)
            self.rawserver.add_task(r)
        except AttributeError:
            pass

    def getResponse(self):
        try:
            return self.response
        except:
            return None

    def Pause(self):
        if not self.storagewrapper:
            return False
        self.unpauseflag.clear()
        self.rawserver.add_task(self.onPause)
        return True

    def onPause(self):
        self.whenpaused = clock()
        if not self.downloader:
            return
        self.downloader.pause(True)
        self.encoder.pause(True)
        self.choker.pause(True)
    
    def Unpause(self):
        self.unpauseflag.set()
        self.rawserver.add_task(self.onUnpause)

    def onUnpause(self):
        if not self.downloader:
            return
        self.downloader.pause(False)
        self.encoder.pause(False)
        self.choker.pause(False)
        if self.rerequest and self.whenpaused and clock()-self.whenpaused > 60:
            self.rerequest.announce(3)      # rerequest automatically if paused for >60 seconds

    def set_super_seed(self):
        try:
            self.superseedflag.set()
            def s(self = self):
                if self.finflag.isSet():
                    self._set_super_seed()
            self.rawserver.add_task(s)
        except AttributeError:
            pass

    def _set_super_seed(self):
        if not self.super_seeding_active:
            self.super_seeding_active = True
            self.errorfunc('        ** SUPER-SEED OPERATION ACTIVE **\n' +
                           '  please set Max uploads so each peer gets 6-8 kB/s')
            def s(self = self):
                self.downloader.set_super_seed()
                self.choker.set_super_seed()
            self.rawserver.add_task(s)
            if self.finflag.isSet():        # mode started when already finished
                def r(self = self):
                    self.rerequest.announce(3)  # so after kicking everyone off, reannounce
                self.rawserver.add_task(r)

    def am_I_finished(self):
        return self.finflag.isSet()

    def get_transfer_stats(self):
        return self.upmeasure.get_total(), self.downmeasure.get_total()
    
