#!/usr/bin/env python
#
# Written by Bram Cohen
# Modified by Cameron Dale
# see LICENSE.txt for license information
#
# $Id: btdownloadheadless.py 63 2007-05-26 21:04:19Z camrdale-guest $

"""A single downloader for debtorrent.

This script implements a command-line client for managing a single 
debtorrent download.

"""

from DebTorrent import PSYCO
if PSYCO.psyco:
    try:
        import psyco
        assert psyco.__version__ >= 0x010100f0
        psyco.full()
    except:
        pass
    
from DebTorrent.download_bt1 import BT1Download, defaults, parse_params, get_usage, get_response
from DebTorrent.RawServer import RawServer, UPnP_ERROR
from random import seed
from socket import error as socketerror
from DebTorrent.bencode import bencode
from DebTorrent.natpunch import UPnP_test
from threading import Event
from os.path import abspath
from sys import argv, stdout
import sys
from sha import sha
from time import strftime
from DebTorrent.clock import clock
from DebTorrent import createPeerID, version
from DebTorrent.ConfigDir import ConfigDir

assert sys.version >= '2', "Install Python 2.0 or greater"
try:
    True
except:
    True = 1
    False = 0

DEBUG = True

PROFILER = False

def hours(n):
    """Formats seconds into a human-readable time.
    
    Formats a given number of seconds into a human-readable time appropriate
    for display to the user.
    
    @type n: C{int}
    @param n: the number of seconds
    @rtype: C{string}
    @return: a displayable representation of the number of seconds
    
    """
    
    if n == 0:
        return 'complete!'
    try:
        n = int(n)
        assert n >= 0 and n < 5184000  # 60 days
    except:
        return '<unknown>'
    m, s = divmod(n, 60)
    h, m = divmod(m, 60)
    if h > 0:
        return '%d hour %02d min %02d sec' % (h, m, s)
    else:
        return '%d min %02d sec' % (m, s)

class HeadlessDisplayer:
    """Displays on the command-line the status of the download.
    
    Implements a displayer for showing in text format the current status
    of the download.
    
    @type done: C{boolean}
    @ivar done: whether the download has completed or not
    @type file: C{string}
    @ivar file: the download file name to display
    @type percentDone: C{string}
    @ivar percentDone: the current percentage of the download that is complete
    @type timeEst: C{string}
    @ivar timeEst: the estimated time remaining
    @type downloadTo: C{string}
    @ivar downloadTo: the location of the download show to the user
    @type downRate: C{string}
    @ivar downRate: the last download rate shown to the user
    @type upRate: C{string}
    @ivar upRate: the last upload rate shown to the user
    @type shareRating: C{string}
    @ivar shareRating: the last sharing information shown to the user
    @type seedStatus: C{string}
    @ivar seedStatus: information on the current status of seeds seen
    @type peerStatus: C{string}
    @ivar peerStatus: information on the current status of peers seen
    @type errors: C{list} of C{string}
    @ivar errors: a list of all the errors that have occurred, in the order
        in which they occurred
    @type last_update_time: C{float}
    @ivar last_update_time: the time in seconds of the last update
    
    """
    
    def __init__(self):
        """Initialize the status of the download."""
        self.done = False
        self.file = ''
        self.percentDone = ''
        self.timeEst = ''
        self.downloadTo = ''
        self.downRate = ''
        self.upRate = ''
        self.shareRating = ''
        self.seedStatus = ''
        self.peerStatus = ''
        self.errors = []
        self.last_update_time = -1

    def finished(self):
        """Concludes the download successfully."""
        self.done = True
        self.percentDone = '100'
        self.timeEst = 'Download Succeeded!'
        self.downRate = ''
        self.display()

    def failed(self):
        """Ends the download when an error has occurred."""
        self.done = True
        self.percentDone = '0'
        self.timeEst = 'Download Failed!'
        self.downRate = ''
        self.display()

    def error(self, errormsg):
        """Displays an error to the user."""
        self.errors.append(errormsg)
        self.display()

    def display(self, dpflag = Event(), fractionDone = None, timeEst = None, 
            downRate = None, upRate = None, activity = None,
            statistics = None,  **kws):
        """Displays the current status of the download.
        
        Writes a status update to standard output.
        
        @type dpflag: a threading event
        @param dpflag: unknown
        @type fractionDone: C{float}
        @param fractionDone: the amount of the download that is complete [0-1]
        @type timeEst: C{int}
        @param timeEst: the number of seconds remaining to complete the download
        @type downRate: C{int} or C{float}
        @param downRate: the current download rate in bytes per second
        @type upRate: C{int} or C{float}
        @param upRate: the current upload rate in bytes per second
        @type activity: C{string}
        @param activity: the currently occurring activity (overwrites the timeEst)
        @type statistics: L{DebTorrent.BT1.Statistics.Statistics}
        @param statistics: various statistics of the current download
        
        """
        
        if self.last_update_time + 0.1 > clock() and fractionDone not in (0.0, 1.0) and activity is not None:
            return
        self.last_update_time = clock()        
        if fractionDone is not None:
            self.percentDone = str(float(int(fractionDone * 1000)) / 10)
        if timeEst is not None:
            self.timeEst = hours(timeEst)
        if activity is not None and not self.done:
            self.timeEst = activity
        if downRate is not None:
            self.downRate = '%.1f kB/s' % (float(downRate) / (1 << 10))
        if upRate is not None:
            self.upRate = '%.1f kB/s' % (float(upRate) / (1 << 10))
        if statistics is not None:
           if (statistics.shareRating < 0) or (statistics.shareRating > 100):
               self.shareRating = 'oo  (%.1f MB up / %.1f MB down)' % (float(statistics.upTotal) / (1<<20), float(statistics.downTotal) / (1<<20))
           else:
               self.shareRating = '%.3f  (%.1f MB up / %.1f MB down)' % (statistics.shareRating, float(statistics.upTotal) / (1<<20), float(statistics.downTotal) / (1<<20))
           if not self.done:
              self.seedStatus = '%d seen now, plus %.3f distributed copies' % (statistics.numSeeds,0.001*int(1000*statistics.numCopies))
           else:
              self.seedStatus = '%d seen recently, plus %.3f distributed copies' % (statistics.numOldSeeds,0.001*int(1000*statistics.numCopies))
           self.peerStatus = '%d seen now, %.1f%% done at %.1f kB/s' % (statistics.numPeers,statistics.percentDone,float(statistics.torrentRate) / (1 << 10))
        print '\n\n\n\n'
        for err in self.errors:
            print 'ERROR:\n' + err + '\n'
        print 'saving:        ', self.file
        print 'percent done:  ', self.percentDone
        print 'time left:     ', self.timeEst
        print 'download to:   ', self.downloadTo
        print 'download rate: ', self.downRate
        print 'upload rate:   ', self.upRate
        print 'share rating:  ', self.shareRating
        print 'seed status:   ', self.seedStatus
        print 'peer status:   ', self.peerStatus
        stdout.flush()
        dpflag.set()        

    def chooseFile(self, default, size, saveas, dir):
        """Sets up the save directory name.
        
        Takes the default and user-provided save directory, saves them, and
        returns the directory to be used for the download.
        
        @type default: C{string}
        @param default: the default save location
        @type size: C{int}
        @param size: the size of the download in bytes
        @type saveas: C{string}
        @param saveas: the user-provided directory to save the download to,
            or an empty string indicating the default should be used
        @type dir: unknown
        @param dir: unknown
        @rtype: C{string}
        @return: the directory to save the download to
        
        """
        
        self.file = '%s (%.1f MB)' % (default, float(size) / (1 << 20))
        if saveas != '':
            default = saveas
        self.downloadTo = abspath(default)
        return default

    def newpath(self, path):
        """Overwrite the previously saved download location."""
        self.downloadTo = path

def run(params):
    """Runs the downloader.
    
    The main function used to create the displayer and start the download.
    
    @type params: C{list} of C{strings}
    @param params: a list of the command-line arguments given to the script
    
    """
    
    h = HeadlessDisplayer()
    while 1:
        configdir = ConfigDir('downloadheadless')
        defaultsToIgnore = ['responsefile', 'url', 'priority']
        configdir.setDefaults(defaults,defaultsToIgnore)
        configdefaults = configdir.loadConfig()
        defaults.append(('save_options',0,
         "whether to save the current options as the new default configuration " +
         "(only for btdownloadheadless.py)"))
        try:
            config = parse_params(params, configdefaults)
        except ValueError, e:
            print 'error: ' + str(e) + '\nrun with no args for parameter explanations'
            break
        if not config:
            print get_usage(defaults, 80, configdefaults)
            break
        if config['save_options']:
            configdir.saveConfig(config)
        configdir.deleteOldCacheData(config['expire_cache_data'])

        myid = createPeerID()
        seed(myid)
        
        doneflag = Event()
        def disp_exception(text):
            """Prints exceptions to standard output."""
            print text
        rawserver = RawServer(doneflag, config['timeout_check_interval'],
                              config['timeout'], ipv6_enable = config['ipv6_enabled'],
                              failfunc = h.failed, errorfunc = disp_exception)
        upnp_type = UPnP_test(config['upnp_nat_access'])
        while True:
            try:
                listen_port = rawserver.find_and_bind(config['minport'], config['maxport'],
                                config['bind'], ipv6_socket_style = config['ipv6_binds_v4'],
                                upnp = upnp_type, randomizer = config['random_port'])
                break
            except socketerror, e:
                if upnp_type and e == UPnP_ERROR:
                    print 'WARNING: COULD NOT FORWARD VIA UPnP'
                    upnp_type = 0
                    continue
                print "error: Couldn't listen - " + str(e)
                h.failed()
                return

        (response, status_priority) = get_response(config['responsefile'], config['url'], config['status_to_download'], h.error)
        if not response:
            break

        infohash = sha(bencode(response['info'])).digest()

        dow = BT1Download(h.display, h.finished, h.error, disp_exception, doneflag,
                        config, response, infohash, myid, rawserver, listen_port,
                        status_priority, configdir)
        
        if not dow.saveAs(h.chooseFile, h.newpath):
            break

        if not dow.initFiles(old_style = True):
            break
        if DEBUG:
            print 'initFiles complete'
        if not dow.startEngine():
            dow.shutdown()
            break
        if DEBUG:
            print 'Engine started'
        dow.startRerequester()
        if DEBUG:
            print 'Rerequester started'
        dow.autoStats()
        if DEBUG:
            print 'autoStats complete'

        if not dow.am_I_finished():
            h.display(activity = 'connecting to peers')
        if DEBUG:
            print 'preparing to listen forever'
        rawserver.listen_forever(dow.getPortHandler())
        h.display(activity = 'shutting down')
        dow.shutdown()
        break
    try:
        rawserver.shutdown()
    except:
        pass
    if not h.done:
        h.failed()

if __name__ == '__main__':
    if argv[1:] == ['--version']:
        print version
        sys.exit(0)

    if PROFILER:
        import profile, pstats
        p = profile.Profile()
        p.runcall(run, argv[1:])
        p.dump_stats('btdownloadheadless.pstat')
#        pstats.Stats(p).strip_dirs().sort_stats('cumulative').print_stats()
        pstats.Stats(p).strip_dirs().sort_stats('time').print_stats()
    else:
        run(argv[1:])
